@@ -1,6 +1,6 @@
 """Convert CellDesigner models to SBML-qual with a rather strict semantics.
 
-Copyright (C) 2019-2020 Sylvain.Soliman@inria.fr
+Copyright (C) 2019-2021 Sylvain.Soliman@inria.fr
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
@@ -29,6 +29,7 @@ from loguru import logger  # type: ignore
 
 import networkx as nx  # type: ignore
 
+from . import bmaExport
 from . import version
 
 NS = {
@@ -54,6 +55,10 @@ Transition = collections.namedtuple(
 )
 
 
+INHIBITION = ("INHIBITION", "UNKNOWN_INHIBITION")
+NEGATIVE = ("INHIBITION", "NEGATIVE_INFLUENCE", "UNKNOWN_INHIBITION")
+
+
 def read_celldesigner(filename: IO):
     """Parse the given file."""
     root = etree.parse(filename).getroot()
@@ -83,18 +88,19 @@ def species_info(model):
         model.findall(
             "./sbml:annotation/cd:extension/"
             + "cd:listOfComplexSpeciesAliases/"
-            + "cd:complexSpeciesAlias[@compartmentAlias]",
+            + "cd:complexSpeciesAlias",
             NS,
         ),
         model.findall(
             "./sbml:annotation/cd:extension/"
             + "cd:listOfSpeciesAliases/"
-            + "cd:speciesAlias[@compartmentAlias]",
+            + "cd:speciesAlias",
             NS,
         ),
     ):
         bound = species.find(".//cd:bounds", NS)
-        if bound is None:
+        in_complex = species.get("complexSpeciesAlias")
+        if bound is None or in_complex is not None:
             continue
         ref_species = species.get("species")
         logger.debug("parsing ref_species: {ref}", ref=ref_species)
@@ -114,6 +120,11 @@ def species_info(model):
         else:
             is_receptor = False
         mods = get_mods(annot.find(".//cd:listOfModifications", NS))
+        activity = annot.find(".//cd:structuralState", NS)
+        if activity is not None:
+            activity = activity.get("structuralState")
+        else:
+            activity = "inactive"
         name = make_name_precise(sbml.get("name"), classtype, mods)
         comp_id = species.get("compartmentAlias")
         if comp_id in compartments:
@@ -121,8 +132,9 @@ def species_info(model):
         else:
             compartment = find_compartment(comp_id, model)
             compartments[comp_id] = compartment
-        nameconv[species.get("id")] = {
-            "activity": get_text(species.find(".//cd:activity", NS), "inactive"),
+        species_id = species.get("id")
+        nameconv[species_id] = {
+            "activity": activity,
             "x": bound.get("x"),
             "y": bound.get("y"),
             "h": bound.get("h"),
@@ -160,6 +172,8 @@ def find_protein_type(annotation, model):
 
 def find_compartment(comp_id, model):
     """Look for the name of the SBML compartment associated to a CD one."""
+    if comp_id is None:
+        return "default_compartment"
     sbml_id = model.find('.//cd:compartmentAlias[@id="' + comp_id + '"]', NS).get(
         "compartment"
     )
@@ -211,7 +225,15 @@ def add_rdf(nameconv, reference: str, new_rdf: Optional[etree.Element]):
         rdfs = new_rdf.find("./rdf:Description", NS)
         if rdfs is None:
             return
-        nameconv[reference]["annotations"].find("./rdf:Description", NS).extend(rdfs[:])
+        logger.debug(
+            "adding annotations from {rdfs} to {reference}",
+            rdfs=rdfs[:],
+            reference=reference,
+        )
+        description = nameconv[reference]["annotations"].find("./rdf:Description", NS)
+        for element in rdfs[:]:
+            if element not in description:
+                description.append(element)
     else:
         nameconv[reference]["annotations"] = new_rdf
 
@@ -314,7 +336,6 @@ def write_qual(
     qlist = etree.SubElement(model, "qual:listOfQualitativeSpecies")
     tlist = etree.SubElement(model, "qual:listOfTransitions")
     graph = nx.DiGraph()
-    fix_all_names(info)
     add_transitions(tlist, info, graph)
     remove_connected_components(tlist, info, graph, remove)
     if sif:
@@ -354,11 +375,16 @@ def remove_connected_components(
 
 
 def write_sif(sbml_filename: str, info, graph: nx.DiGraph):
-    """Write a SIF file with influences."""
+    """Write a SIF file with influences.
+
+    http://www.cbmc.it/fastcent/doc/SifFormat.htm
+    """
     with open(sbml_filename[:-4] + "sif", "w", encoding="utf-8", newline="") as f:
+        print("# Generated by CaSQ v{version}".format(version=version), file=f)
         with open(
             sbml_filename[:-5] + "_raw.sif", "w", encoding="utf-8", newline=""
         ) as fraw:
+            print("# Generated by CaSQ v{version}".format(version=version), file=fraw)
             for source, target, sign in graph.edges.data("sign"):
                 print(
                     source.replace(" ", "_"),
@@ -374,7 +400,7 @@ def write_sif(sbml_filename: str, info, graph: nx.DiGraph):
                 )
 
 
-def simplify_model(info):
+def simplify_model(info, upstream, downstream):
     """Clean the model w.r.t. some active/inactive species."""
     multispecies = delete_complexes_and_store_multispecies(info)
     # pylint: disable=too-many-nested-blocks
@@ -414,6 +440,8 @@ def simplify_model(info):
                             key=key,
                         )
                         del info[val]
+    fix_all_names(info)
+    restrict_model(info, upstream, downstream)
 
 
 def delete_complexes_and_store_multispecies(info):
@@ -423,6 +451,8 @@ def delete_complexes_and_store_multispecies(info):
     proteins that bind to form a complex.
     """
     multispecies = {}  # type: Dict[str, List[str]]
+    duplicate_nodes = {}
+    replacements = {}
     # we have to create the list since info will change during iteration
     for key, value in list(info.items()):
         if key.startswith("__"):
@@ -430,6 +460,27 @@ def delete_complexes_and_store_multispecies(info):
             logger.debug("deleting complex: {cplx} value: {val}", cplx=key, val=value)
             if len(value) > 1:
                 multispecies[key] = value
+        # merge nodes that have the same reference species
+        elif (
+            value["ref_species"] in duplicate_nodes
+            and key in info
+            and duplicate_nodes[value["ref_species"]] in info
+        ):
+            into = duplicate_nodes[value["ref_species"]]
+            logger.debug(
+                "merging {key} into {into} for {ref} ({name})",
+                key=key,
+                into=into,
+                ref=value["ref_species"],
+                name=value["name"],
+            )
+            # put our annotations with those of into
+            add_rdf(info, into, info[key]["annotations"])
+            # put our transitions with those of into
+            info[into]["transitions"].extend(info[key]["transitions"])
+            # replace key in other transitions with into
+            replacements[key] = into
+            del info[key]
         # delete receptors that only contribute to their complex
         elif value["receptor"] and not value["transitions"]:
             logger.debug("{key} is a RECEPTOR (and an input)", key=key)
@@ -461,6 +512,8 @@ def delete_complexes_and_store_multispecies(info):
                         and active2 == key
                         and not info[reac1]["receptor"]
                         and not info[reac2]["receptor"]
+                        and not info[reac1]["transitions"]
+                        and not info[reac2]["transitions"]
                     ):
                         logger.debug(
                             "deleting {reac1} and {reac2} for complex {key}",
@@ -470,13 +523,78 @@ def delete_complexes_and_store_multispecies(info):
                         )
                         add_rdf(info, key, info[reac1]["annotations"])
                         add_rdf(info, key, info[reac2]["annotations"])
+
                         info[key]["transitions"].extend(info[reac1]["transitions"])
                         info[key]["transitions"].extend(info[reac2]["transitions"])
+                        if info[reac1]["ref_species"] in duplicate_nodes:
+                            duplicate_nodes[info[reac1]["ref_species"]] = key
+                        if info[reac2]["ref_species"] in duplicate_nodes:
+                            duplicate_nodes[info[reac2]["ref_species"]] = key
                         del info[reac1]
                         del info[reac2]
+                        # info[key]["transitions"].remove(trans)
+        else:
+            duplicate_nodes[value["ref_species"]] = key
+    replace_in_transitions(info, replacements)
     return multispecies
 
 
+def restrict_model(info, upstream, downstream):
+    """Only keep species upstream/downstream of some list of species."""
+    name_to_ids = {v["name"]: k for (k, v) in info.items()}
+    for name in upstream + downstream:
+        if name not in name_to_ids:
+            logger.error(name + " was not found, maybe it is ambiguous…")
+
+    if upstream == [] and downstream == []:
+        return
+    graph = nx.DiGraph()
+    for species, data in info.items():
+        graph.add_node(species)
+        for trans in data["transitions"]:
+            for val in trans.reactants:
+                graph.add_edge(val, species)
+            for (_modtype, modlist) in trans.modifiers:
+                for val in modlist.split(","):
+                    graph.add_edge(val, species)
+    keep = set()
+    for dnname in downstream:
+        dn = name_to_ids[dnname]
+        keep |= {
+            elt for succl in nx.dfs_successors(graph, dn).values() for elt in succl
+        } | {dn}
+    ggraph = graph.reverse()
+    for upname in upstream:
+        up = name_to_ids[upname]
+        keep |= {
+            elt for succl in nx.dfs_successors(ggraph, up).values() for elt in succl
+        } | {up}
+    for species in list(info.keys()):
+        if species not in keep:
+            del info[species]
+
+
+def replace_in_transitions(info, replacements):
+    """Change transitions in info to reflect replacements."""
+    for _species, data in info.items():
+        for trans in data["transitions"]:
+            for val in trans.reactants.copy():
+                if val in replacements:
+                    trans.reactants.append(replacements[val])
+                    trans.reactants.remove(val)
+            for (modtype, mod_list) in trans.modifiers.copy():
+                mlist = mod_list.split(",")
+                changed = False
+                for val in mlist:
+                    if val in replacements:
+                        mlist.append(replacements[val])
+                        mlist.remove(val)
+                        changed = True
+                if changed:
+                    trans.modifiers.append((modtype, ",".join(mlist)))
+                    trans.modifiers.remove((modtype, mod_list))
+
+
 def get_active(val, info):
     """Find who val activates."""
     active = None
@@ -505,11 +623,26 @@ def fix_all_names(info):
     ambiguous_name = {
         key: count_names[value["name"]] > 1 for key, value in info.items()
     }
+    namedict = {}
     for species, data in info.items():
-        data["name"] = fix_name(
-            data["name"], ambiguous_name[species], data["compartment"]
-        )
-        data["function"] = data["name"]
+        name = fix_name(data["name"], ambiguous_name[species], data["compartment"])
+        activity = data["activity"]
+        if ambiguous_name[species]:
+            if name in namedict:
+                other_id, other_activity = namedict[name]
+                if activity == "active":
+                    # FIXME what if name_active in namedict?
+                    name = name + "_active"
+                elif other_activity == "active":
+                    info[other_id]["name"] = name + "_active"
+                    info[other_id]["function"] = name + "_active"
+                else:
+                    # FIXME don't know what to do
+                    # print(f"active is {activity}, other was {other_activity}")
+                    pass
+            namedict[name] = (species, activity)
+        data["name"] = name
+        data["function"] = name
 
 
 def add_qual_species(layout: etree.Element, qlist: etree.Element, info):
@@ -680,20 +813,23 @@ def add_function(func: etree.Element, transitions: List[Transition], known: List
         activators = [
             modifier
             for (modtype, modifier) in reaction.modifiers
-            if modtype not in ("INHIBITION", "BOOLEAN_LOGIC_GATE_AND")
+            if modtype
+            not in ("INHIBITION", "UNKNOWN_INHIBITION", "BOOLEAN_LOGIC_GATE_AND")
             and modifier in known
             and modifier not in reactants
         ]
         inhibitors = [
             modifier
             for (modtype, modifier) in reaction.modifiers
-            if modtype == "INHIBITION" and modifier in known
+            if modtype in INHIBITION and modifier in known
         ]
         # this should only appear when species is of type PHENOTYPE otherwise
-        # non-SBGN compliant
+        # non-SBGN compliant, and there should be a single reactant and no inhibitors
         # just swap reactants and inhibitors, there should not be any activator
-        if reaction.type == "NEGATIVE_INFLUENCE":
+        if reaction.type in NEGATIVE:
             reactants, inhibitors = inhibitors, reactants
+            if activators or reactants:
+                logger.debug("non-SBGN direct inhibition encountered")
         # create and node if necessary
         if len(reactants) + len(inhibitors) > 1 or (
             activators and (reactants or inhibitors)
@@ -749,13 +885,13 @@ def add_inputs(
         for modtype, modifier in chain(
             enumerate(reaction.reactants), reaction.modifiers
         ):
-            if modtype == "INHIBITION":
+            if modtype in INHIBITION:
                 sign = "negative"
             else:
                 sign = "positive"
             # this should only appear when species is of type PHENOTYPE
             # otherwise non-SBGN compliant
-            if reaction.type == "NEGATIVE_INFLUENCE":
+            if reaction.type in NEGATIVE:
                 sign = negate(sign)
             if (modifier, sign) not in modifiers and modifier in known:
                 modifiers.append((modifier, sign))
@@ -854,7 +990,7 @@ def main():
         version="%(prog)s v{version}".format(version=version),
     )
     parser.add_argument(
-        "-d", "--debug", action="store_true", help="Display a lot of debug information"
+        "-D", "--debug", action="store_true", help="Display a lot of debug information"
     )
     parser.add_argument(
         "-c",
@@ -862,6 +998,19 @@ def main():
         action="store_true",
         help="Store the species information in a separate CSV file",
     )
+    parser.add_argument(
+        "-b",
+        "--bma",
+        action="store_true",
+        help="Output to BMA json format",
+    )
+    parser.add_argument(
+        "-g",
+        "--granularity",
+        type=int,
+        default="1",
+        help="When exporting to BMA, use this granularity",
+    )
     parser.add_argument(
         "-s",
         "--sif",
@@ -878,6 +1027,42 @@ def main():
         size is smaller than S.
         A negative S leads to keep only the biggest(s) connected component(s)""",
     )
+    if sys.version_info >= (3, 8, 0):
+        parser.add_argument(
+            "-u",
+            "--upstream",
+            action="extend",
+            nargs="*",
+            type=str,
+            default=[],
+            help="Only species upstream of this/these species will be kept",
+        )
+        parser.add_argument(
+            "-d",
+            "--downstream",
+            action="extend",
+            nargs="*",
+            type=str,
+            default=[],
+            help="Only species downstream of this/these species will be kept",
+        )
+    else:
+        parser.add_argument(
+            "-u",
+            "--upstream",
+            action="append",
+            type=str,
+            default=[],
+            help="Only species upstream of this/these species will be kept",
+        )
+        parser.add_argument(
+            "-d",
+            "--downstream",
+            action="append",
+            type=str,
+            default=[],
+            help="Only species downstream of this/these species will be kept",
+        )
     parser.add_argument(
         "infile",
         type=argparse.FileType("r", encoding="utf-8"),
@@ -885,17 +1070,22 @@ def main():
         default=sys.stdin,
         help="CellDesigner File",
     )
-    parser.add_argument("outfile", nargs="?", default=sys.stdout, help="SBML-Qual File")
+    parser.add_argument(
+        "outfile", nargs="?", default=sys.stdout, help="SBML-Qual/BMA json File"
+    )
     args = parser.parse_args()
 
     if not args.debug:
         logger.disable("casq")
     logger.debug("parsing {fname}…", fname=args.infile.name)
     info, width, height = read_celldesigner(args.infile)
-    simplify_model(info)
+    simplify_model(info, args.upstream, args.downstream)
     if args.infile != sys.stdin and args.outfile == sys.stdout:
         args.outfile = os.path.splitext(args.infile.name)[0] + ".sbml"
-    write_qual(args.outfile, info, width, height, remove=args.remove, sif=args.sif)
+    if args.bma:
+        bmaExport.write_bma(args.outfile, info, args.granularity)
+    else:
+        write_qual(args.outfile, info, width, height, remove=args.remove, sif=args.sif)
     if args.csv and args.outfile != sys.stdout:
         write_csv(args.outfile, info)
 
